from typing import Sequence, cast
from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

BASIC_PATTERNS = "basic_patterns"
PREFIXED_PATTERNS = "prefixed_patterns"
SUFFIXED_PATTERNS = "suffixed_patterns"
COMMON_PATTERN = "suffixed_patterns"
WITH_NULL = "with_null"

DATA = pd.DataFrame(
    {
        BASIC_PATTERNS: ["abc", "def", "ghi"],
        PREFIXED_PATTERNS: ["foo_abc", "foo_def", "foo_ghi"],
        SUFFIXED_PATTERNS: ["abc_foo", "def_foo", "ghi_foo"],
        COMMON_PATTERN: ["abc_foo", "def_foo", "ghi_foo"],
        WITH_NULL: ["ba", None, "ab"],
    }
)

SUPPORTED_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]


@parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
def test_basic_success(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToMatchLikePatternList(
        column=PREFIXED_PATTERNS,
        like_pattern_list=["foo%"],
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
def test_basic_failure(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToMatchLikePatternList(
        column=BASIC_PATTERNS,
        like_pattern_list=["xyz%"],
    )
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig(), RedshiftDatasourceTestConfig()],
    data=DATA,
)
def test_complete_results_failure(batch_for_datasource: Batch) -> None:
    ABOUT_TWO_THIRDS = pytest.approx(2 / 3 * 100)
    expectation = gxe.ExpectColumnValuesToMatchLikePatternList(
        column=BASIC_PATTERNS,
        like_pattern_list=["%b%"],
    )
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    json_dict = result.to_json_dict()
    result_dict = json_dict.get("result")

    assert isinstance(result_dict, dict)
    assert not result.success
    assert "IS NOT NULL AND basic_patterns NOT LIKE '%b%'" in cast(
        "str", result_dict.get("unexpected_index_query")
    )
    assert result.to_json_dict().get("result") == {
        "element_count": 3,
        "unexpected_count": 2,
        "unexpected_percent": ABOUT_TWO_THIRDS,
        "partial_unexpected_list": ["def", "ghi"],
        "missing_count": 0,
        "missing_percent": 0.0,
        "unexpected_percent_total": ABOUT_TWO_THIRDS,
        "unexpected_percent_nonmissing": ABOUT_TWO_THIRDS,
        "partial_unexpected_counts": [
            {"value": "def", "count": 1},
            {"value": "ghi", "count": 1},
        ],
        "unexpected_list": ["def", "ghi"],
        "unexpected_index_query": ANY,
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=COMMON_PATTERN, like_pattern_list=["%foo"], match_on="any"
            ),
            id="match_all",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS, like_pattern_list=["abc", "def", "ghi"], match_on="any"
            ),
            id="multiple_patterns",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS, like_pattern_list=["a%", "d%", "g%"], match_on="any"
            ),
            id="multiple_patterns_with_prefix",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=COMMON_PATTERN, like_pattern_list=["%foo"], match_on="any"
            ),
            id="match_all",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=PREFIXED_PATTERNS, like_pattern_list=["foo%"], match_on="any"
            ),
            id="prefixed_pattern",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=SUFFIXED_PATTERNS, like_pattern_list=["%foo"], match_on="any"
            ),
            id="suffixed_pattern",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=PREFIXED_PATTERNS,
                like_pattern_list=["foo%", "bar%"],
                match_on="any",
            ),
            id="matches_one_of",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS, like_pattern_list=["%b%"], match_on="any", mostly=0.3
            ),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig(), RedshiftDatasourceTestConfig()],
    data=DATA,
)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToMatchLikePatternList,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS,
                like_pattern_list=["%xyz%"],
                match_on="any",
            ),
            id="no_matches",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS,
                like_pattern_list=["%b%"],
                match_on="any",
                mostly=0.4,
            ),
            id="mostly_threshold_not_met",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=PREFIXED_PATTERNS,
                like_pattern_list=["foo%", "bar%"],
                match_on="all",
            ),
            id="does_not_match_all",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[PostgreSQLDatasourceTestConfig(), RedshiftDatasourceTestConfig()],
    data=DATA,
)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToMatchLikePatternList,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS,
                like_pattern_list=["[adg]%"],
                match_on="any",
            ),
        ),
        pytest.param(
            gxe.ExpectColumnValuesToMatchLikePatternList(
                column=BASIC_PATTERNS,
                like_pattern_list=["[a]%", "[d]%", "[g]%"],
                match_on="any",
            ),
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=[MSSQLDatasourceTestConfig()], data=DATA)
def test_msql_fancy_syntax(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToMatchLikePatternList,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param("any", True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=SUPPORTED_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_match_on_(
    batch_for_datasource: Batch, suite_param_value: str, expected_result: bool
) -> None:
    suite_param_key = "test_expect_column_values_to_match_like_pattern_list"

    expectation = gxe.ExpectColumnValuesToMatchLikePatternList(
        column=BASIC_PATTERNS,
        like_pattern_list=["abc", "def", "ghi"],
        match_on={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

import pytest

from great_expectations.experimental.metric_repository.data_store import DataStore
from great_expectations.experimental.metric_repository.metric_repository import (
    MetricRepository,
)
from great_expectations.experimental.metric_repository.metrics import MetricRun


@pytest.fixture
def mock_data_store():
    return Mock(autospec=DataStore)


@pytest.fixture
def mock_metric_run():
    return Mock(autospec=MetricRun)


@pytest.mark.unit
def test_add_metric_run(mock_data_store: DataStore, mock_metric_run: MetricRun):
    metric_repository = MetricRepository(data_store=mock_data_store)

    metric_repository.add_metric_run(metric_run=mock_metric_run)

    mock_data_store.add.assert_called_once_with(value=mock_metric_run)  # type: ignore[attr-defined] # FIXME CoP



import os
import sys
from pathlib import Path

import pytest

# Add the project root to the Python path so we can import from setup.py
project_root = Path(__file__).parent.parent.parent
sys.path.insert(0, str(project_root))

from setup import get_extras_require, parse_requirements


class TestSetupIntegration:
    @pytest.mark.filesystem
    def test_parse_main_requirements_txt(self):
        """Test parsing the main requirements.txt file."""
        requirements_file = project_root / "requirements.txt"

        assert requirements_file.exists(), f"requirements.txt not found at {requirements_file}"

        requirements = parse_requirements(requirements_file)

        assert len(requirements) > 0, "requirements.txt should contain at least one requirement"
        assert all(req.strip() for req in requirements), "No requirements should be empty strings"
        assert not any(req.startswith("#") for req in requirements), (
            "No comment lines should be in the result"
        )

        print(f"Found {len(requirements)} requirements in requirements.txt")
        print("First few requirements:", requirements[:5])

    @pytest.mark.filesystem
    def test_parse_dev_requirements_files(self):
        """Test parsing all dev requirements files in the reqs/ directory."""
        reqs_dir = project_root / "reqs"

        assert reqs_dir.exists(), f"reqs directory not found at {reqs_dir}"
        requirements_files = list(reqs_dir.glob("requirements-dev-*.txt"))
        assert len(requirements_files) > 0, "Should find at least one dev requirements file"

        for req_file in requirements_files:
            requirements = parse_requirements(req_file)

            assert all(isinstance(req, str) for req in requirements), (
                f"All requirements in {req_file.name} should be strings"
            )
            assert all(req.strip() for req in requirements), (
                f"No requirements in {req_file.name} should be empty strings"
            )
            assert not any(req.startswith("#") for req in requirements), (
                f"No comment lines should be in the result for {req_file.name}"
            )

            print(f"Parsed {req_file.name}: {len(requirements)} requirements")

    @pytest.mark.filesystem
    def test_get_extras_require_functionality(self):
        """Test that get_extras_require() works end-to-end."""
        # Change to project root for relative path resolution
        original_cwd = Path.cwd()
        try:
            os.chdir(project_root)

            extras = get_extras_require()

            assert len(extras) > 0, "Should have at least one extra requirement set"
            for key, requirements in extras.items():
                assert isinstance(requirements, list), f"Extra '{key}' should be a list"
                assert all(isinstance(req, str) for req in requirements), (
                    f"All requirements for extra '{key}' should be strings"
                )
                assert all(req.strip() for req in requirements), (
                    f"No requirements for extra '{key}' should be empty strings"
                )

            print(f"Found {len(extras)} extra requirement sets")
            print("Extra names:", list(extras.keys()))

        finally:
            os.chdir(original_cwd)



from __future__ import annotations

import logging
from dataclasses import asdict, dataclass
from functools import cached_property
from typing import Any, Type

import sqlalchemy as sa
from sqlalchemy.pool import Pool, QueuePool

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ConnectionDetails:
    connection_string: str

    @cached_property
    def dialect(self) -> str:
        return sa.engine.make_url(self.connection_string).get_backend_name()


@dataclass(frozen=True)
class PoolConfig:
    # The sqlalchemy connection pool class to use. In general we want to use QueuePool
    poolclass: Type[Pool]
    # The number of connections to keep in the pool
    pool_size: int
    # If all pool connections are used, we can create an additional max_overflow connections
    # When returning connections to the pool, if the pool is full, additional connections will
    # be discarded. This is specific for a QueuePool.
    max_overflow: int
    # The number of seconds a connection can be open for before we recycle it and create a new one.
    pool_recycle: int
    # Number of seconds to wait before giving up on getting a connection from the pool
    pool_timeout: int
    # Test connection liveness on checkout
    pool_pre_ping: bool


class SessionSQLEngineManager:
    POOL_CONFIG = PoolConfig(
        poolclass=QueuePool,
        pool_size=2,
        max_overflow=3,
        pool_recycle=5400,  # 1.5 hours
        pool_timeout=30,  # 30 seconds
        pool_pre_ping=True,
    )

    def __init__(self):
        # It's ok to use ConnectionDetails as the key since that contains all the unique
        # information needed to create an engine. If we allowed POOL_CONFIG to be configurable
        # we'd need to incorporate that into the key.
        self._engine_cache: dict[ConnectionDetails, sa.engine.Engine] = {}

    def get_engine(
        self,
        connection_details: ConnectionDetails,
    ) -> sa.engine.Engine:
        cache_key = connection_details
        if cache_key not in self._engine_cache:
            logger.info(f"Cache miss for engine: {cache_key}. Creating new engine.")
            engine_kwargs = asdict(self.POOL_CONFIG)
            logger.info(
                f"Creating engine for {connection_details.dialect} with settings: {engine_kwargs}"
            )
            self._engine_cache[cache_key] = sa.create_engine(
                connection_details.connection_string, **engine_kwargs
            )
        else:
            logger.info(f"Cache hit for engine: {cache_key}")
        return self._engine_cache[cache_key]

    def dispose_all_engines(self):
        logger.info("Disposing all cached SQLAlchemy engines.")
        for key, engine in self._engine_cache.items():
            logger.info(f"Disposing engine: {key}")
            try:
                engine.dispose()
            except Exception:
                logger.exception(f"Error disposing engine '{key}'")
        self._engine_cache.clear()

    def get_all_pool_statistics(
        self,
    ) -> dict[ConnectionDetails, dict[str, Any]]:
        stats: dict[ConnectionDetails, dict[str, Any]] = {}
        for key, engine in self._engine_cache.items():
            try:
                pool = engine.pool
                if isinstance(pool, QueuePool):
                    stats[key] = {
                        "size": pool.size(),
                        "checked_in": pool.checkedin(),
                        "overflow": pool.overflow(),
                        "checked_out": pool.checkedout(),
                    }
                else:
                    logger.warning(
                        f"Pool for engine {key} is not a QueuePool. It is a {type(pool)}."
                    )
                    stats[key] = {
                        "type": f"{type(pool)}",
                        "status": f"{pool.status()}",
                    }
            except Exception as e:
                logger.exception(f"Error getting pool status for engine '{key}'")
                stats[key] = {"error": str(e)}
        return stats



from __future__ import annotations

from enum import Enum


class PrimaryColors(str, Enum):
    ORANGE = "#FF6310"
    COAL_GRAY = "#404041"


class SecondaryColors(str, Enum):
    MIDNIGHT_BLUE = "#272763"
    ROYAL_BLUE = "#223F99"
    TURQUOISE_BLUE = "#4DC0B4"
    LEAF_GREEN = "#B6C647"
    GOLD_YELLOW = "#F3C62D"
    POMEGRANATE_PINK = "#E2226B"
    LAVENDER_PURPLE = "#BC87E6"


class TintsAndShades(str, Enum):
    ORANGE_100 = "#803208"
    ORANGE_90 = "#A03e0A"
    ORANGE_80 = "#C04B0C"
    ORANGE_70 = "#df570e"
    ORANGE_60 = "#ff6310"
    ORANGE_50 = "#ff7f3b"
    ORANGE_40 = "#ff9b67"
    ORANGE_30 = "#ffb892"
    ORANGE_20 = "#ffd4be"
    ORANGE_10 = "#FFF0E9"
    YELLOW_100 = "#534207"
    YELLOW_90 = "#7b6311"
    YELLOW_80 = "#a3841a"
    YELLOW_70 = "#cba524"
    YELLOW_60 = "#f3c62d"
    YELLOW_50 = "#f5d053"
    YELLOW_40 = "#f7da78"
    YELLOW_30 = "#fae59e"
    YELLOW_20 = "#fcefc3"
    YELLOW_10 = "#fef9e9"
    PINK_100 = "#70062e"
    PINK_90 = "#960f42"
    PINK_80 = "#bc1957"
    PINK_70 = "#e2226b"
    PINK_60 = "#e64281"
    PINK_50 = "#ea6396"
    PINK_40 = "#ef83ac"
    PINK_30 = "#f3a3c1"
    PINK_20 = "#f7c4d6"
    PINK_10 = "#fbe4ec"
    GRAY_100 = "#000000"
    GRAY_90 = "#404041"
    GRAY_80 = "#58595B"
    GRAY_70 = "#6d6e70"
    GRAY_60 = "#808184"
    GRAY_50 = "#929497"
    GRAY_40 = "#a6a8ab"
    GRAY_30 = "#bbbdbf"
    GRAY_20 = "#d0d2d3"
    GRAY_10 = "#e6e7e8"
    MIDNIGHT_BLUE_100 = "#141432"
    MIDNIGHT_BLUE_90 = "#1e1e48"
    MIDNIGHT_BLUE_80 = "#272763"
    MIDNIGHT_BLUE_70 = "#424277"
    MIDNIGHT_BLUE_60 = "#5d5d8a"
    MIDNIGHT_BLUE_50 = "#78789e"
    MIDNIGHT_BLUE_40 = "#9494b1"
    MIDNIGHT_BLUE_30 = "#afafc5"
    MIDNIGHT_BLUE_20 = "#cacad8"
    MIDNIGHT_BLUE_10 = "#e5e5ec"
    ROYAL_BLUE_100 = "#091a4c"
    ROYAL_BLUE_90 = "#0f235f"
    ROYAL_BLUE_80 = "#162072"
    ROYAL_BLUE_70 = "#1c3686"
    ROYAL_BLUE_60 = "#223f99"
    ROYAL_BLUE_50 = "#4860ab"
    ROYAL_BLUE_40 = "#6e81bc"
    ROYAL_BLUE_30 = "#95a3ce"
    ROYAL_BLUE_20 = "#bbc4df"
    ROYAL_BLUE_10 = "#e1e5f1"
    TURQUOISE_BLUE_100 = "#144944"
    TURQUOISE_BLUE_90 = "#226760"
    TURQUOISE_BLUE_80 = "#31847c"
    TURQUOISE_BLUE_70 = "#3fa298"
    TURQUOISE_BLUE_60 = "#4dc0b4"
    TURQUOISE_BLUE_50 = "#6ccbc1"
    TURQUOISE_BLUE_40 = "#8bd6ce"
    TURQUOISE_BLUE_30 = "#aae1da"
    TURQUOISE_BLUE_20 = "#c9ece7"
    TURQUOISE_BLUE_10 = "#e8f7f4"
    GREEN_100 = "#454c14"
    GREEN_90 = "#616a21"
    GREEN_80 = "#7e892e"
    GREEN_70 = "#9aa83a"
    GREEN_60 = "#b6c647"
    GREEN_50 = "#c3d067"
    GREEN_40 = "#cfda86"
    GREEN_30 = "#dce3a6"
    GREEN_20 = "#e8edc5"
    GREEN_10 = "#f5f7e5"
    PURPLE_100 = "#3b254d"
    PURPLE_90 = "#5b3e73"
    PURPLE_80 = "#7c569a"
    PURPLE_70 = "#9c6fc0"
    PURPLE_60 = "#bc87e6"
    PURPLE_50 = "#c79aea"
    PURPLE_40 = "#d2adee"
    PURPLE_30 = "#dcc1f2"
    PURPLE_20 = "#e7d4f6"
    PURPLE_10 = "#f2e7fa"


class ColorPalettes(Enum):
    CATEGORY_5 = [
        PrimaryColors.ORANGE,
        SecondaryColors.ROYAL_BLUE,
        SecondaryColors.TURQUOISE_BLUE,
        SecondaryColors.LEAF_GREEN,
        SecondaryColors.LAVENDER_PURPLE,
    ]
    CATEGORY_7 = [
        PrimaryColors.ORANGE,
        SecondaryColors.ROYAL_BLUE,
        SecondaryColors.TURQUOISE_BLUE,
        SecondaryColors.LEAF_GREEN,
        SecondaryColors.GOLD_YELLOW,
        SecondaryColors.POMEGRANATE_PINK,
        SecondaryColors.LAVENDER_PURPLE,
    ]
    DIVERGING_7 = [
        TintsAndShades.GREEN_50,
        TintsAndShades.GREEN_30,
        TintsAndShades.GREEN_10,
        "#f1f1f1",
        TintsAndShades.PINK_10,
        TintsAndShades.PINK_30,
        TintsAndShades.PINK_50,
    ]
    HEATMAP_6 = [
        TintsAndShades.TURQUOISE_BLUE_10,
        TintsAndShades.TURQUOISE_BLUE_30,
        TintsAndShades.TURQUOISE_BLUE_50,
        TintsAndShades.MIDNIGHT_BLUE_50,
        TintsAndShades.MIDNIGHT_BLUE_70,
        TintsAndShades.MIDNIGHT_BLUE_90,
    ]
    ORDINAL_5 = [
        TintsAndShades.ORANGE_90,
        TintsAndShades.ORANGE_70,
        TintsAndShades.ORANGE_50,
        TintsAndShades.YELLOW_50,
        TintsAndShades.YELLOW_30,
    ]



from typing import List

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

snowflake_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []

partition_data = [
    IntegrationTestFixture(
        name="partition_data_on_whole_table_snowflake",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_whole_table.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/snowflake_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.SNOWFLAKE],
    ),
    IntegrationTestFixture(
        name="partition_data_on_datetime_snowflake",
        user_flow_script="tests/integration/db/test_sql_data_partitioned_on_datetime_and_day_part.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        data_dir="tests/test_sets/taxi_yellow_tripdata_samples/",
        util_script="tests/test_utils.py",
        other_files=(
            (
                "tests/integration/fixtures/partition_and_sample_data/snowflake_connection_string.yml",
                "connection_string.yml",
            ),
        ),
        backend_dependencies=[BackendDependencies.SNOWFLAKE],
    ),
]

sample_data: List[IntegrationTestFixture] = []

snowflake_integration_tests += connecting_to_your_data
snowflake_integration_tests += partition_data
snowflake_integration_tests += sample_data



from __future__ import annotations

import logging
import re
import warnings
from typing import TYPE_CHECKING, Callable, ClassVar, List, Optional, Type

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_spec import GCSBatchSpec, PathBatchSpec
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    FilePathDataConnector,
    MissingFilePathTemplateMapFnError,
    sanitize_prefix_for_gcs_and_s3,
)

if TYPE_CHECKING:
    from great_expectations.alias_types import PathStr
    from great_expectations.compatibility import google
    from great_expectations.core.batch import LegacyBatchDefinition


logger = logging.getLogger(__name__)


class _GCSOptions(pydantic.BaseModel):
    gcs_prefix: str = ""
    gcs_delimiter: str = "/"
    gcs_max_results: int = 1000
    gcs_recursive_file_discovery: bool = False


class GoogleCloudStorageDataConnector(FilePathDataConnector):
    """Extension of FilePathDataConnector used to connect to Google Cloud Storage (GCS).

    Args:
        datasource_name: The name of the Datasource associated with this DataConnector instance
        data_asset_name: The name of the DataAsset using this DataConnector instance
        gcs_client: Reference to instantiated Google Cloud Storage client handle
        bucket_or_name (str): bucket name for Google Cloud Storage
        prefix (str): GCS prefix
        delimiter (str): GCS delimiter
        max_results (int): max blob filepaths to return
        recursive_file_discovery (bool): Flag to indicate if files should be searched recursively from subfolders
        file_path_template_map_fn: Format function mapping path to fully-qualified resource on GCS
        whole_directory_path_override: If present, treat entire directory as single Asset
    """  # noqa: E501 # FIXME CoP

    asset_level_option_keys: ClassVar[tuple[str, ...]] = (
        "gcs_prefix",
        "gcs_delimiter",
        "gcs_max_results",
        "gcs_recursive_file_discovery",
    )
    asset_options_type: ClassVar[Type[_GCSOptions]] = _GCSOptions

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_asset_name: str,
        gcs_client: google.Client,
        bucket_or_name: str,
        prefix: str = "",
        delimiter: str = "/",
        max_results: Optional[int] = None,
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> None:
        self._gcs_client: google.Client = gcs_client

        self._bucket_or_name = bucket_or_name

        self._prefix: str = prefix
        self._sanitized_prefix: str = sanitize_prefix_for_gcs_and_s3(text=prefix)

        self._delimiter = delimiter
        self._max_results = max_results

        self._recursive_file_discovery = recursive_file_discovery

        super().__init__(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_data_connector(  # noqa: PLR0913 # FIXME CoP
        cls,
        datasource_name: str,
        data_asset_name: str,
        gcs_client: google.Client,
        bucket_or_name: str,
        prefix: str = "",
        delimiter: str = "/",
        max_results: Optional[int] = None,
        recursive_file_discovery: bool = False,
        file_path_template_map_fn: Optional[Callable] = None,
        whole_directory_path_override: PathStr | None = None,
    ) -> GoogleCloudStorageDataConnector:
        """Builds "GoogleCloudStorageDataConnector", which links named DataAsset to Google Cloud Storage.

        Args:
            datasource_name: The name of the Datasource associated with this "GoogleCloudStorageDataConnector" instance
            data_asset_name: The name of the DataAsset using this "GoogleCloudStorageDataConnector" instance
            gcs_client: Reference to instantiated Google Cloud Storage client handle
            bucket_or_name: bucket name for Google Cloud Storage
            prefix: GCS prefix
            delimiter: GCS delimiter
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders
            max_results: max blob filepaths to return
            file_path_template_map_fn: Format function mapping path to fully-qualified resource on GCS
            whole_directory_path_override: If present, treat entire directory as single Asset

        Returns:
            Instantiated "GoogleCloudStorageDataConnector" object
        """  # noqa: E501 # FIXME CoP
        return GoogleCloudStorageDataConnector(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            gcs_client=gcs_client,
            bucket_or_name=bucket_or_name,
            prefix=prefix,
            delimiter=delimiter,
            max_results=max_results,
            recursive_file_discovery=recursive_file_discovery,
            file_path_template_map_fn=file_path_template_map_fn,
            whole_directory_path_override=whole_directory_path_override,
        )

    @classmethod
    def build_test_connection_error_message(
        cls,
        data_asset_name: str,
        bucket_or_name: str,
        prefix: str = "",
        delimiter: str = "/",
        recursive_file_discovery: bool = False,
    ) -> str:
        """Builds helpful error message for reporting issues when linking named DataAsset to Google Cloud Storage.

        Args:
            data_asset_name: The name of the DataAsset using this "GoogleCloudStorageDataConnector" instance
            bucket_or_name: bucket name for Google Cloud Storage
            prefix: GCS prefix
            delimiter: GCS delimiter
            recursive_file_discovery: Flag to indicate if files should be searched recursively from subfolders

        Returns:
            Customized error message
        """  # noqa: E501 # FIXME CoP
        test_connection_error_message_template: str = 'No file in bucket "{bucket_or_name}" with prefix "{prefix}" and recursive file discovery set to "{recursive_file_discovery}" found using delimiter "{delimiter}" for DataAsset "{data_asset_name}".'  # noqa: E501 # FIXME CoP
        return test_connection_error_message_template.format(
            **{
                "data_asset_name": data_asset_name,
                "bucket_or_name": bucket_or_name,
                "prefix": prefix,
                "delimiter": delimiter,
                "recursive_file_discovery": recursive_file_discovery,
            }
        )

    @override
    def build_batch_spec(self, batch_definition: LegacyBatchDefinition) -> GCSBatchSpec:
        """
        Build BatchSpec from batch_definition by calling DataConnector's build_batch_spec function.

        Args:
            batch_definition (LegacyBatchDefinition): to be used to build batch_spec

        Returns:
            BatchSpec built from batch_definition
        """
        batch_spec: PathBatchSpec = super().build_batch_spec(batch_definition=batch_definition)
        return GCSBatchSpec(batch_spec)

    # Interface Method
    @override
    def get_data_references(self) -> List[str]:
        query_options: dict = {
            "bucket_or_name": self._bucket_or_name,
            "prefix": self._sanitized_prefix,
            "delimiter": self._delimiter,
            "max_results": self._max_results,
        }
        path_list: List[str] = list_gcs_keys(
            gcs_client=self._gcs_client,
            query_options=query_options,
            recursive=self._recursive_file_discovery,
        )
        return path_list

    # Interface Method
    @override
    def _get_full_file_path(self, path: str) -> str:
        # If the path is already a fully qualified GCS URL (starts with gs://), return it as-is
        # This handles the case of whole_directory_path_override which is already fully qualified
        if path.startswith("gs://"):
            return path

        if self._file_path_template_map_fn is None:
            raise MissingFilePathTemplateMapFnError()

        template_arguments = {
            "bucket_or_name": self._bucket_or_name,
            "path": path,
        }

        return self._file_path_template_map_fn(**template_arguments)

    @override
    def _preprocess_batching_regex(self, regex: re.Pattern) -> re.Pattern:
        regex = re.compile(f"{re.escape(self._sanitized_prefix)}{regex.pattern}")
        return super()._preprocess_batching_regex(regex=regex)


def list_gcs_keys(
    gcs_client,
    query_options: dict,
    recursive: bool = False,
) -> List[str]:
    """
    Utilizes the GCS connection object to retrieve blob names based on user-provided criteria.

    For InferredAssetGCSDataConnector, we take `bucket_or_name` and `prefix` and search for files using RegEx at and below the level
    specified by those parameters. However, for ConfiguredAssetGCSDataConnector, we take `bucket_or_name` and `prefix` and
    search for files using RegEx only at the level specified by that bucket and prefix.

    This restriction for the ConfiguredAssetGCSDataConnector is needed because paths on GCS are comprised not only the leaf file name
    but the full path that includes both the prefix and the file name. Otherwise, in the situations where multiple data assets
    share levels of a directory tree, matching files to data assets will not be possible due to the path ambiguity.

    Please note that the SDK's `list_blobs` method takes in a `delimiter` key that drastically alters the traversal of a given bucket:
        - If a delimiter is not set (default), the traversal is recursive and the output will contain all blobs in the current directory
          as well as those in any nested directories.
        - If a delimiter is set, the traversal will continue until that value is seen; as the default is "/", traversal will be scoped
          within the current directory and end before visiting nested directories.

    In order to provide users with finer control of their config while also ensuring output that is in line with the `recursive` arg,
    we deem it appropriate to manually override the value of the delimiter only in cases where it is absolutely necessary.

    Args:
        gcs_client (storage.Client): GCS connnection object responsible for accessing bucket
        query_options (dict): GCS query attributes ("bucket_or_name", "prefix", "delimiter", "max_results")
        recursive (bool): True for InferredAssetGCSDataConnector and False for ConfiguredAssetGCSDataConnector (see above)

    Returns:
        List of keys representing GCS file paths (as filtered by the `query_options` dict)
    """  # noqa: E501 # FIXME CoP
    # Delimiter determines whether or not traversal of bucket is recursive
    # Manually set to appropriate default if not already set by user
    delimiter = query_options["delimiter"]
    if delimiter is None and not recursive:
        warnings.warn(
            'In order to access blobs with a ConfiguredAssetGCSDataConnector, \
            or with a Fluent datasource without enabling recursive file discovery, \
            the delimiter that has been passed to gcs_options in your config cannot be empty; \
            please note that the value is being set to the default "/" in order to work with the Google SDK.'  # noqa: E501 # FIXME CoP
        )
        query_options["delimiter"] = "/"
    elif delimiter is not None and recursive:
        warnings.warn(
            "In order to access blobs with an InferredAssetGCSDataConnector, \
            or enabling recursive file discovery with a Fluent datasource, \
            the delimiter that has been passed to gcs_options in your config must be empty; \
            please note that the value is being set to None in order to work with the Google SDK."
        )
        query_options["delimiter"] = None

    keys: List[str] = []
    for blob in gcs_client.list_blobs(**query_options):
        name: str = blob.name
        if name.endswith("/"):  # GCS includes directories in blob output
            continue

        keys.append(name)

    return keys



from __future__ import annotations

import os.path
import pathlib
from pprint import pformat as pf
from typing import Final, List

import pytest
import requirements as rp

IGNORE_PINS: Final[set[str]] = {"mypy", "ruff", "pytest"}


def collect_requirements_files() -> List[pathlib.Path]:
    project_root = pathlib.Path(__file__).parents[1]
    assert project_root.exists()
    reqs_dir = project_root.joinpath("reqs")
    assert reqs_dir.exists()

    pattern = "requirements*.txt"
    requirement_files = list(project_root.glob(pattern)) + list(reqs_dir.glob(pattern))
    return requirement_files


def parse_requirements_files_to_strings(
    files: list[pathlib.Path],
) -> dict[str, set[str]]:
    """Parse requirements files to dict.

    dict of the form {"filename": {"package_name_with_specs_as_str"}}

    Args:
        files: Paths to files that should be parsed
    """

    req_set_dict: dict[str, set[str]] = {}
    for req_file in files:
        abs_path = req_file.absolute().as_posix()
        key = abs_path.rsplit(os.path.sep, 1)[-1]
        with open(req_file) as f:
            req_set_dict[key] = {
                f"{line.name}{','.join([''.join(spec) for spec in line.specs])}"
                for line in rp.parse(f)
                if line.specs
            }

    return req_set_dict


def parse_requirements_files_to_specs(
    files: list[pathlib.Path],
) -> dict[str, dict[str | None, list[str]]]:
    """Parse requirements files to dict.

    dict of the form {"filename": {"package_name_with_specs_as_str"}}

    Args:
        files: Paths to files that should be parsed
    """

    req_set_dict: dict[str, dict[str | None, list[str]]] = {}
    for req_file in files:
        abs_path = req_file.absolute().as_posix()
        key = abs_path.rsplit(os.path.sep, 1)[-1]
        with open(req_file) as f:
            req_set_dict[key] = {line.name: line.specs for line in rp.parse(f) if line.specs}  # type: ignore[misc] # FIXME CoP

    return req_set_dict


@pytest.mark.unit
def test_requirements_files():
    """requirements.txt should be a subset of requirements-dev.txt"""

    req_files = collect_requirements_files()
    req_set_dict = parse_requirements_files_to_strings(files=req_files)

    assert req_set_dict["requirements.txt"] <= req_set_dict["requirements-dev.txt"]

    assert (
        req_set_dict["requirements-dev-arrow.txt"]
        | req_set_dict["requirements-dev-contrib.txt"]
        | req_set_dict["requirements-dev-lite.txt"]
        | req_set_dict["requirements-dev-api-docs-test.txt"]
        == req_set_dict["requirements-dev-test.txt"]
    )

    assert (
        req_set_dict["requirements-dev-lite.txt"] & req_set_dict["requirements-dev-spark.txt"]
        == set()
    )

    assert (
        req_set_dict["requirements-dev-spark.txt"]
        & req_set_dict["requirements-dev-sqlalchemy.txt"]
        & req_set_dict["requirements-dev-azure.txt"]
        == set()
    )

    assert (
        req_set_dict["requirements-dev-lite.txt"] & req_set_dict["requirements-dev-contrib.txt"]
        == set()
    )

    assert (
        req_set_dict["requirements-dev-lite.txt"]
        | req_set_dict["requirements-dev-athena.txt"]
        | req_set_dict["requirements-dev-bigquery.txt"]
        | req_set_dict["requirements-dev-dremio.txt"]
        | req_set_dict["requirements-dev-mssql.txt"]
        | req_set_dict["requirements-dev-mysql.txt"]
        | req_set_dict["requirements-dev-postgresql.txt"]
        | req_set_dict["requirements-dev-gx-redshift.txt"]
        | req_set_dict["requirements-dev-snowflake.txt"]
        | req_set_dict["requirements-dev-teradata.txt"]
        | req_set_dict["requirements-dev-clickhouse.txt"]
        | req_set_dict["requirements-dev-trino.txt"]
        | req_set_dict["requirements-dev-hive.txt"]
        | req_set_dict["requirements-dev-vertica.txt"]
        | req_set_dict["requirements-dev-sqlalchemy1.txt"]
    ) == req_set_dict["requirements-dev-sqlalchemy.txt"]

    assert (
        req_set_dict["requirements.txt"]
        | req_set_dict["requirements-dev-contrib.txt"]
        | req_set_dict["requirements-dev-test.txt"]
        | req_set_dict["requirements-dev-sqlalchemy.txt"]
        | req_set_dict["requirements-dev-arrow.txt"]
        | req_set_dict["requirements-dev-azure.txt"]
        | req_set_dict["requirements-dev-excel.txt"]
        | req_set_dict["requirements-dev-pagerduty.txt"]
        | req_set_dict["requirements-dev-spark.txt"]
    ) == req_set_dict["requirements-dev.txt"]

    assert req_set_dict["requirements-dev.txt"] - (
        req_set_dict["requirements.txt"]
        | req_set_dict["requirements-dev-lite.txt"]
        | req_set_dict["requirements-dev-contrib.txt"]
        | req_set_dict["requirements-dev-test.txt"]
        | req_set_dict["requirements-dev-spark.txt"]
        | req_set_dict["requirements-dev-sqlalchemy1.txt"]
        | req_set_dict["requirements-dev-sqlalchemy.txt"]
        | req_set_dict["requirements-dev-arrow.txt"]
        | req_set_dict["requirements-dev-athena.txt"]
        | req_set_dict["requirements-dev-azure.txt"]
        | req_set_dict["requirements-dev-bigquery.txt"]
        | req_set_dict["requirements-dev-dremio.txt"]
        | req_set_dict["requirements-dev-excel.txt"]
        | req_set_dict["requirements-dev-mssql.txt"]
        | req_set_dict["requirements-dev-mysql.txt"]
        | req_set_dict["requirements-dev-pagerduty.txt"]
        | req_set_dict["requirements-dev-postgresql.txt"]
        | req_set_dict["requirements-dev-gx-redshift.txt"]
        | req_set_dict["requirements-dev-snowflake.txt"]
        | req_set_dict["requirements-dev-teradata.txt"]
        | req_set_dict["requirements-dev-clickhouse.txt"]
        | req_set_dict["requirements-dev-trino.txt"]
        | req_set_dict["requirements-dev-vertica.txt"]
    ) <= {"numpy>=1.21.0", "scipy>=1.7.0"}


@pytest.mark.unit
def test_polish_and_ratchet_pins_and_upper_bounds():
    """What does this test do and why?

    We would like to reduce the number of pins and upper bounds on dependencies
    so that we can stay up to date with our dependencies where possible.
    """
    req_files = collect_requirements_files()
    req_set_dict = parse_requirements_files_to_specs(files=req_files)
    # don't care about type-checking requirements
    req_set_dict.pop("requirements-types.txt")
    print(f"Requirement Groups:\n{pf(req_set_dict, depth=1)}")

    packages_with_pins_or_upper_bounds = set()

    for req_file, package_specs in req_set_dict.items():
        for package, specs in package_specs.items():
            if package in IGNORE_PINS:
                continue
            for spec in specs:
                if spec[0] in ("<", "<=", "=="):
                    packages_with_pins_or_upper_bounds.add(
                        (
                            req_file,
                            package,
                            tuple(sorted(specs, key=lambda s: (s[0], s[1]))),
                        )
                    )

    sorted_packages_with_pins_or_upper_bounds = sorted(
        list(packages_with_pins_or_upper_bounds), key=lambda p: (p[0], p[1])
    )

    # Polish and ratchet this number down as low as possible
    assert len(sorted_packages_with_pins_or_upper_bounds) == 38
    assert set(sorted_packages_with_pins_or_upper_bounds) == {
        (
            "requirements-dev-api-docs-test.txt",
            "docstring-parser",
            (("==", "0.16"),),
        ),
        ("requirements-dev-athena.txt", "pyathena", (("<", "3"), (">=", "2.0.0"))),
        ("requirements-dev-contrib.txt", "adr-tools-python", (("==", "1.0.3"),)),
        ("requirements-dev-dremio.txt", "sqlalchemy-dremio", (("==", "1.2.1"),)),
        ("requirements-dev-excel.txt", "xlrd", (("<", "2.0.0"), (">=", "1.1.0"))),
        ("requirements-dev-lite.txt", "moto", (("<", "5.0"), (">=", "4.2.13"))),
        ("requirements-dev-pagerduty.txt", "pypd", (("==", "1.1.0"),)),
        (
            "requirements-dev-spark.txt",
            "pyspark",
            (("<", "4.0"), (">=", "2.3.2")),
        ),
        ("requirements-dev-snowflake.txt", "pandas", (("<", "2.2.0"),)),
        ("requirements-dev-sqlalchemy.txt", "moto", (("<", "5.0"), (">=", "4.2.13"))),
        ("requirements-dev-sqlalchemy.txt", "pandas", (("<", "2.2.0"),)),
        (
            "requirements-dev-sqlalchemy.txt",
            "pyathena",
            (("<", "3"), (">=", "2.0.0")),
        ),
        ("requirements-dev-sqlalchemy.txt", "sqlalchemy", (("<", "2.0.0"),)),
        (
            "requirements-dev-sqlalchemy.txt",
            "sqlalchemy-dremio",
            (("==", "1.2.1"),),
        ),
        (
            "requirements-dev-sqlalchemy.txt",
            "teradatasqlalchemy",
            (("==", "17.0.0.5"),),
        ),
        ("requirements-dev-sqlalchemy1.txt", "sqlalchemy", (("<", "2.0.0"),)),
        (
            "requirements-dev-teradata.txt",
            "teradatasqlalchemy",
            (("==", "17.0.0.5"),),
        ),
        ("requirements-dev-test.txt", "adr-tools-python", (("==", "1.0.3"),)),
        ("requirements-dev-test.txt", "docstring-parser", (("==", "0.16"),)),
        ("requirements-dev-test.txt", "moto", (("<", "5.0"), (">=", "4.2.13"))),
        ("requirements-dev.txt", "adr-tools-python", (("==", "1.0.3"),)),
        ("requirements-dev.txt", "altair", (("<", "5.0.0"), (">=", "4.2.1"))),
        ("requirements-dev.txt", "docstring-parser", (("==", "0.16"),)),
        ("requirements-dev.txt", "marshmallow", (("<", "4.0.0"), (">=", "3.7.1"))),
        ("requirements-dev.txt", "moto", (("<", "5.0"), (">=", "4.2.13"))),
        ("requirements-dev.txt", "pandas", (("<", "2.2.0"),)),
        ("requirements-dev.txt", "posthog", (("<", "6"), (">", "3"))),
        ("requirements-dev.txt", "pyathena", (("<", "3"), (">=", "2.0.0"))),
        (
            "requirements-dev.txt",
            "pyspark",
            (("<", "4.0"), (">=", "2.3.2")),
        ),
        ("requirements-dev.txt", "pypd", (("==", "1.1.0"),)),
        ("requirements-dev.txt", "sqlalchemy", (("<", "2.0.0"),)),
        ("requirements-dev.txt", "sqlalchemy-dremio", (("==", "1.2.1"),)),
        ("requirements-dev.txt", "teradatasqlalchemy", (("==", "17.0.0.5"),)),
        ("requirements-dev.txt", "xlrd", (("<", "2.0.0"), (">=", "1.1.0"))),
        ("requirements.txt", "altair", (("<", "5.0.0"), (">=", "4.2.1"))),
        ("requirements.txt", "marshmallow", (("<", "4.0.0"), (">=", "3.7.1"))),
        ("requirements.txt", "pandas", (("<", "2.2"),)),
        ("requirements.txt", "posthog", (("<", "6"), (">", "3"))),
    }



import pandas as pd
import pytest
import sqlalchemy.types as sqltypes
from packaging import version

import great_expectations.expectations as gxe
from great_expectations.compatibility.aws import REDSHIFT_TYPES
from great_expectations.compatibility.databricks import DATABRICKS_TYPES
from great_expectations.compatibility.postgresql import POSTGRESQL_TYPES
from great_expectations.compatibility.snowflake import SNOWFLAKE_TYPES
from great_expectations.compatibility.sqlalchemy import (
    sqlalchemy as sa,
)
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import (
    DatabricksDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
)

INTEGER_COLUMN = "integers"
INTEGER_AND_NULL_COLUMN = "integers_and_nulls"
STRING_COLUMN = "strings"


DATA = pd.DataFrame(
    {
        INTEGER_COLUMN: [1, 2, 3, 4, 5],
        INTEGER_AND_NULL_COLUMN: [1, 2, 3, 4, None],
        STRING_COLUMN: ["a", "b", "c", "d", "e"],
    },
    dtype="object",
)

PASSING_DATA_SOURCES_EXCEPT_DATA_FRAMES = [
    ds
    for ds in ALL_DATA_SOURCES
    if not isinstance(
        ds,
        (
            PandasDataFrameDatasourceTestConfig,
            SnowflakeDatasourceTestConfig,
            DatabricksDatasourceTestConfig,
        ),
    )
]


@parameterize_batch_for_data_sources(
    data_source_configs=PASSING_DATA_SOURCES_EXCEPT_DATA_FRAMES, data=DATA
)
def test_success_complete(batch_for_datasource: Batch) -> None:
    type_list = [
        "INTEGER",
        "Integer",
        "int",
        "int64",
        "int32",
        "IntegerType",
        "_CUSTOM_DECIMAL",
    ]
    expectation = gxe.ExpectColumnValuesToBeInTypeList(column=INTEGER_COLUMN, type_list=type_list)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    result_dict = result.to_json_dict()["result"]

    assert result.success
    assert isinstance(result_dict, dict)
    assert result_dict["observed_value"] in type_list


@parameterize_batch_for_data_sources(
    data_source_configs=[PandasDataFrameDatasourceTestConfig()], data=DATA
)
def test_success_complete_pandas(batch_for_datasource: Batch) -> None:
    type_list = ["INTEGER", "int", "int64", "int32", "IntegerType"]
    expectation = gxe.ExpectColumnValuesToBeInTypeList(column=INTEGER_COLUMN, type_list=type_list)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)

    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column=INTEGER_COLUMN, type_list=["int"]),
            id="integer_types",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column=INTEGER_AND_NULL_COLUMN, type_list=["int", "float64"], mostly=0.8
            ),
            id="mostly",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column=STRING_COLUMN, type_list=["str"]),
            id="string_types",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeInTypeList,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column=INTEGER_COLUMN, type_list=["str"]),
            id="wrong_type",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToBeInTypeList,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="CHARACTER", type_list=["CHARACTER", "VARCHAR(1)"]
            ),
            id="CHARACTER",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="DEC", type_list=["DEC", "DECIMAL", "DECIMAL(38, 0)"]
            ),
            id="DEC",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="FIXED", type_list=["FIXED", "DECIMAL", "DECIMAL(38, 0)"]
            ),
            id="FIXED",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="GEOGRAPHY", type_list=["GEOGRAPHY"]),
            id="GEOGRAPHY",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="GEOMETRY", type_list=["GEOMETRY"]),
            id="GEOMETRY",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="NUMBER",
                type_list=[
                    "NUMBER",
                    "DECIMAL",
                    "NUMERIC",
                    "DECIMAL(38, 0)",  # 38, 0 is the default precision and scale
                ],
            ),
            id="NUMBER",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="STRING", type_list=["STRING", "VARCHAR", "VARCHAR(16777216)"]
            ),
            id="STRING",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="TEXT", type_list=["TEXT", "VARCHAR", "VARCHAR(16777216)"]
            ),
            id="TEXT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="TIMESTAMP_LTZ", type_list=["TIMESTAMP_LTZ"]
            ),
            id="TIMESTAMP_LTZ",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="TIMESTAMP_NTZ", type_list=["TIMESTAMP_NTZ"]
            ),
            id="TIMESTAMP_NTZ",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TIMESTAMP_TZ", type_list=["TIMESTAMP_TZ"]),
            id="TIMESTAMP_TZ",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="VARBINARY", type_list=["VARBINARY", "BINARY"]
            ),
            id="VARBINARY",
        ),
        # INT , INTEGER , BIGINT , SMALLINT , TINYINT , BYTEINT are Synonymous with NUMBER,
        # except that precision and scale cannot be specified (i.e. always defaults to NUMBER(38, 0)
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="BYTEINT", type_list=["DECIMAL(38, 0)"]),
            id="BYTEINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TINYINT", type_list=["DECIMAL(38, 0)"]),
            id="TINYINT",
        ),
        # Complex data types which are not hashable by testing framework currently
        # pytest.param(
        #     gxe.ExpectColumnValuesToBeInTypeList(
        #         column="VARIANT", type_list=["VARIANT"]
        #     ),
        #     id="VARIANT",
        # ),
        # pytest.param(
        #     gxe.ExpectColumnValuesToBeInTypeList(column="OBJECT", type_list=["OBJECT"]),
        #     id="OBJECT",
        # ),
        # pytest.param(
        #     gxe.ExpectColumnValuesToBeInTypeList(column="ARRAY", type_list=["ARRAY"]),
        #     id="ARRAY",
        # ),
        # These sqlachemy types map to _CUSTOM_* types in snowflake-sqlalchemy
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="_CUSTOM_Date", type_list=["DATE"]),
            id="_CUSTOM_Date",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="_CUSTOM_DateTime", type_list=["TIMESTAMP_NTZ"]
            ),
            id="_CUSTOM_DateTime",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="_CUSTOM_Time", type_list=["TIME"]),
            id="_CUSTOM_Time",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="_CUSTOM_Float", type_list=["FLOAT"]),
            id="_CUSTOM_Float",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="_CUSTOM_DECIMAL", type_list=["INTEGER", "DECIMAL(38, 0)"]
            ),
            id="_CUSTOM_DECIMAL",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[
        SnowflakeDatasourceTestConfig(
            column_types={
                "ARRAY": SNOWFLAKE_TYPES.ARRAY,
                "BYTEINT": SNOWFLAKE_TYPES.BYTEINT,
                "CHARACTER": SNOWFLAKE_TYPES.CHARACTER,
                "DEC": SNOWFLAKE_TYPES.DEC,
                "FIXED": SNOWFLAKE_TYPES.FIXED,
                "GEOGRAPHY": SNOWFLAKE_TYPES.GEOGRAPHY,
                "GEOMETRY": SNOWFLAKE_TYPES.GEOMETRY,
                "NUMBER": SNOWFLAKE_TYPES.NUMBER,
                "OBJECT": SNOWFLAKE_TYPES.OBJECT,
                "STRING": SNOWFLAKE_TYPES.STRING,
                "TEXT": SNOWFLAKE_TYPES.TEXT,
                "TIMESTAMP_LTZ": SNOWFLAKE_TYPES.TIMESTAMP_LTZ,
                "TIMESTAMP_NTZ": SNOWFLAKE_TYPES.TIMESTAMP_NTZ,
                "TIMESTAMP_TZ": SNOWFLAKE_TYPES.TIMESTAMP_TZ,
                "TINYINT": SNOWFLAKE_TYPES.TINYINT,
                "VARBINARY": SNOWFLAKE_TYPES.VARBINARY,
                "VARIANT": SNOWFLAKE_TYPES.VARIANT,
                # These sqlachemy types map to _CUSTOM_* types in snowflake-sqlalchemy
                "_CUSTOM_Date": sqltypes.Date,
                "_CUSTOM_DateTime": sqltypes.DateTime,
                "_CUSTOM_Time": sqltypes.Time,
                "_CUSTOM_Float": sqltypes.Float,
                "_CUSTOM_DECIMAL": sqltypes.INTEGER,
            }
        )
    ],
    data=pd.DataFrame(
        {
            "BYTEINT": [1, 2, 3],
            "CHARACTER": ["a", "b", "c"],
            "DEC": [1.0, 2.0, 3.0],
            "FIXED": [1.0, 2.0, 3.0],
            "GEOGRAPHY": ["POINT(1 1)", "POINT(2 2)", "POINT(3 3)"],
            "GEOMETRY": ["POINT(1 1)", "POINT(2 2)", "POINT(3 3)"],
            "NUMBER": [1, 2, 3],
            "STRING": ["a", "b", "c"],
            "TEXT": ["a", "b", "c"],
            "TIMESTAMP_LTZ": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "TIMESTAMP_NTZ": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "TIMESTAMP_TZ": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "TINYINT": [1, 2, 3],
            "VARBINARY": [b"1", b"2", b"3"],
            # Complex data types which are not hashable by testing framework currently
            # "ARRAY": pd.Series([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype="object"),
            # "OBJECT": [{"a": 1}, {"b": 2}, {"c": 3}],
            # "VARIANT": [{"a": 1}, {"b": 2}, {"c": 3}],
            # These sqlachemy types map to _CUSTOM_* types in snowflake-sqlalchemy
            "_CUSTOM_Date": [
                # Date in isoformat
                "2021-01-01",
                "2021-01-02",
                "2021-01-03",
            ],
            "_CUSTOM_DateTime": [
                # isoformat with microseconds
                "2021-01-01 00:00:00.000000",
                "2021-01-02 00:00:00.000000",
                "2021-01-03 00:00:00.000000",
            ],
            "_CUSTOM_Time": [
                "00:00:00.878281",
                "01:00:00.000000",
                "00:10:43.000000",
            ],
            "_CUSTOM_Float": [1.0, 2.0, 3.0],
            "_CUSTOM_DECIMAL": [1, 2, 3],
        },
        dtype="object",
    ),
)
def test_success_complete_snowflake(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeInTypeList
) -> None:
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    result_dict = result.to_json_dict()["result"]

    assert result.success
    assert isinstance(result_dict, dict)
    assert isinstance(result_dict["observed_value"], str)
    assert isinstance(expectation.type_list, list)
    assert result_dict["observed_value"] in expectation.type_list


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="CHAR", type_list=["CHAR", "CHAR(1)"]),
            id="CHAR",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TEXT", type_list=["TEXT"]),
            id="TEXT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="INTEGER", type_list=["INTEGER"]),
            id="INTEGER",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="SMALLINT", type_list=["SMALLINT"]),
            id="SMALLINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="BIGINT", type_list=["BIGINT"]),
            id="BIGINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="TIMESTAMP", type_list=["TIMESTAMP", "TIMESTAMP WITHOUT TIME ZONE"]
            ),
            id="TIMESTAMP",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="DATE", type_list=["DATE"]),
            id="DATE",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="DOUBLE_PRECISION", type_list=["DOUBLE PRECISION"]
            ),
            id="DOUBLE_PRECISION",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="BOOLEAN", type_list=["BOOLEAN"]),
            id="BOOLEAN",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="NUMERIC", type_list=["NUMERIC"]),
            id="NUMERIC",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[
        PostgreSQLDatasourceTestConfig(
            column_types={
                "CHAR": POSTGRESQL_TYPES.CHAR,
                "TEXT": POSTGRESQL_TYPES.TEXT,
                "INTEGER": POSTGRESQL_TYPES.INTEGER,
                "SMALLINT": POSTGRESQL_TYPES.SMALLINT,
                "BIGINT": POSTGRESQL_TYPES.BIGINT,
                "TIMESTAMP": POSTGRESQL_TYPES.TIMESTAMP,
                "DATE": POSTGRESQL_TYPES.DATE,
                "DOUBLE_PRECISION": POSTGRESQL_TYPES.DOUBLE_PRECISION,
                "BOOLEAN": POSTGRESQL_TYPES.BOOLEAN,
                "NUMERIC": POSTGRESQL_TYPES.NUMERIC,
            }
        ),
    ],
    data=pd.DataFrame(
        {
            "CHAR": ["a", "b", "c"],
            "TEXT": ["a", "b", "c"],
            "INTEGER": [1, 2, 3],
            "SMALLINT": [1, 2, 3],
            "BIGINT": [1, 2, 3],
            "TIMESTAMP": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "DATE": [
                # Date in isoformat
                "2021-01-01",
                "2021-01-02",
                "2021-01-03",
            ],
            "DOUBLE_PRECISION": [1.0, 2.0, 3.0],
            "BOOLEAN": [False, False, True],
            "NUMERIC": [1, 2, 3],
        },
        dtype="object",
    ),
)
def test_success_complete_postgres(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeInTypeList
) -> None:
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    result_dict = result.to_json_dict()["result"]

    assert result.success
    assert isinstance(result_dict, dict)
    assert isinstance(result_dict["observed_value"], str)
    assert isinstance(expectation.type_list, list)
    assert result_dict["observed_value"] in expectation.type_list


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="STRING", type_list=["STRING"]),
            id="STRING",
        ),
        # SqlA Text gets converted to Databricks STRING
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TEXT", type_list=["STRING"]),
            id="TEXT",
        ),
        # SqlA UNICODE gets converted to Databricks STRING
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="UNICODE", type_list=["STRING"]),
            id="UNICODE",
        ),
        # SqlA UNICODE_TEXT gets converted to Databricks STRING
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="UNICODE_TEXT", type_list=["STRING"]),
            id="UNICODE_TEXT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="BOOLEAN", type_list=["BOOLEAN"]),
            id="BOOLEAN",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="DECIMAL", type_list=["DECIMAL", "DECIMAL(10, 0)"]
            ),
            id="DECIMAL",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="DATE", type_list=["DATE"]),
            id="DATE",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TIMESTAMP", type_list=["TIMESTAMP"]),
            id="TIMESTAMP",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="TIMESTAMP_NTZ", type_list=["TIMESTAMP_NTZ"]
            ),
            id="TIMESTAMP_NTZ",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="FLOAT", type_list=["FLOAT"]),
            id="FLOAT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="INT", type_list=["INT"]),
            id="INT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="TINYINT", type_list=["TINYINT"]),
            id="TINYINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="DECIMAL", type_list=["DECIMAL", "DECIMAL(10, 0)"]
            ),
            id="DECIMAL",
        ),
        # SqlA Time gets converted to Databricks STRING,
        # but is not supported by our testing framework
        # pytest.param(
        #     gxe.ExpectColumnValuesToBeInTypeList(column="TIME", type_list=["STRING"]),
        #     id="TIME",
        # ),
        # SqlA UUID gets converted to Databricks STRING,
        # but is not supported by our testing framework.
        # pytest.param(
        #     gxe.ExpectColumnValuesToBeInTypeList(column="UUID", type_list=["STRING"]),
        #     id="UUID",
        # )
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[
        DatabricksDatasourceTestConfig(
            column_types={
                "STRING": DATABRICKS_TYPES.STRING,
                "TEXT": sqltypes.Text,
                "UNICODE": sqltypes.Unicode,
                "UNICODE_TEXT": sqltypes.UnicodeText,
                "BIGINT": sqltypes.BigInteger,
                "BOOLEAN": sqltypes.BOOLEAN,
                "DATE": sqltypes.DATE,
                "TIMESTAMP_NTZ": DATABRICKS_TYPES.TIMESTAMP_NTZ,
                "TIMESTAMP": DATABRICKS_TYPES.TIMESTAMP,
                "FLOAT": sqltypes.Float,
                "INT": sqltypes.Integer,
                "DECIMAL": sqltypes.Numeric,
                "SMALLINT": sqltypes.SmallInteger,
                "TINYINT": DATABRICKS_TYPES.TINYINT,
                # "TIME": sqltypes.Time,
                # "UUID": sqltypes.UUID,
            }
        )
    ],
    data=pd.DataFrame(
        {
            "STRING": ["a", "b", "c"],
            "TEXT": ["a", "b", "c"],
            "UNICODE": ["\u00e9", "\u00e9", "\u00e9"],
            "UNICODE_TEXT": ["a", "b", "c"],
            "BIGINT": [1111, 2222, 3333],
            "BOOLEAN": [True, True, False],
            "DATE": [
                "2021-01-01",
                "2021-01-02",
                "2021-01-03",
            ],
            "TIMESTAMP_NTZ": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "TIMESTAMP": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "DOUBLE": [1.0, 2.0, 3.0],
            "FLOAT": [1.0, 2.0, 3.0],
            "INT": [1, 2, 3],
            "DECIMAL": [1.1, 2.2, 3.3],
            "SMALLINT": [1, 2, 3],
            # "TIME": [
            #     sa.Time("22:17:33.123456"),
            #     sa.Time("22:17:33.123456"),
            #     sa.Time("22:17:33.123456"),
            # ],
            # "UUID": [
            #      uuid.UUID("905993ea-f50e-4284-bea0-5be3f0ed7031"),
            #      uuid.UUID("9406b631-fa2f-41cf-b666-f9a2ac3118c1"),
            #      uuid.UUID("47538f05-32e3-4594-80e2-0b3b33257ae7")
            #  ],
        },
        dtype="object",
    ),
)
def test_success_complete_databricks(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeInTypeList
) -> None:
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    result_dict = result.to_json_dict()["result"]

    assert result.success
    assert isinstance(result_dict, dict)
    assert isinstance(result_dict["observed_value"], str)
    assert isinstance(expectation.type_list, list)
    assert result_dict["observed_value"] in expectation.type_list


if version.parse(sa.__version__) >= version.parse("2.0.0"):
    # Note: why not use pytest.skip?
    # the import of `sqltypes.Double` is only possible in sqlalchemy >= 2.0.0
    # the import is done as part of the instantiation of the test, which includes
    # processing the pytest.skip() statement. This way, we skip the instantiation
    # of the test entirely.
    @pytest.mark.parametrize(
        "expectation",
        [
            pytest.param(
                gxe.ExpectColumnValuesToBeInTypeList(
                    column="DOUBLE", type_list=["DOUBLE", "FLOAT"]
                ),
                id="DOUBLE",
            )
        ],
    )
    @parameterize_batch_for_data_sources(
        data_source_configs=[
            DatabricksDatasourceTestConfig(
                column_types={
                    "DOUBLE": sqltypes.Double,
                }
            )
        ],
        data=pd.DataFrame(
            {
                "DOUBLE": [1.0, 2.0, 3.0],
            },
            dtype="object",
        ),
    )
    def test_success_complete_databricks_double_type_only(
        batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeInTypeList
    ) -> None:
        """What does this test and why?

        Databricks mostly uses SqlA types directly, but the double type is
        only available after sqlalchemy 2.0. We therefore split up the test
        into 2 parts, with this test being skipped if the SA version is too low.
        """
        result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
        result_dict = result.to_json_dict()["result"]

        assert result.success
        assert isinstance(result_dict, dict)
        assert isinstance(result_dict["observed_value"], str)
        assert isinstance(expectation.type_list, list)
        assert result_dict["observed_value"] in expectation.type_list


# Redshift will lowercase column names so we make the column names here all lowercase
# These means we aren't case insensitive in this expectation.
@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="char", type_list=["CHAR", "CHAR(1)"]),
            id="CHAR",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="text", type_list=["VARCHAR"]),
            id="TEXT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="integer", type_list=["INTEGER"]),
            id="INTEGER",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="smallint", type_list=["SMALLINT"]),
            id="SMALLINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="bigint", type_list=["BIGINT"]),
            id="BIGINT",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="timestamp", type_list=["TIMESTAMP"]),
            id="TIMESTAMP",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="date", type_list=["DATE"]),
            id="DATE",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(
                column="double_precision", type_list=["DOUBLE_PRECISION"]
            ),
            id="DOUBLE_PRECISION",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="boolean", type_list=["BOOLEAN"]),
            id="BOOLEAN",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToBeInTypeList(column="numeric", type_list=["DECIMAL"]),
            id="NUMERIC",
        ),
    ],
)
@parameterize_batch_for_data_sources(
    data_source_configs=[
        RedshiftDatasourceTestConfig(
            column_types={
                "char": REDSHIFT_TYPES.CHAR,
                "text": REDSHIFT_TYPES.VARCHAR,
                "integer": REDSHIFT_TYPES.INTEGER,
                "smallint": REDSHIFT_TYPES.SMALLINT,
                "bigint": REDSHIFT_TYPES.BIGINT,
                "timestamp": REDSHIFT_TYPES.TIMESTAMP,
                "date": REDSHIFT_TYPES.DATE,
                "double_precision": REDSHIFT_TYPES.DOUBLE_PRECISION,
                "boolean": REDSHIFT_TYPES.BOOLEAN,
                "numeric": REDSHIFT_TYPES.DECIMAL,
            }
        ),
    ],
    data=pd.DataFrame(
        {
            "char": ["a", "b", "c"],
            "text": ["a", "b", "c"],
            "integer": [1, 2, 3],
            "smallint": [1, 2, 3],
            "bigint": [1, 2, 3],
            "timestamp": [
                "2021-01-01 00:00:00",
                "2021-01-02 00:00:00",
                "2021-01-03 00:00:00",
            ],
            "date": [
                # Date in isoformat
                "2021-01-01",
                "2021-01-02",
                "2021-01-03",
            ],
            "double_precision": [1.0, 2.0, 3.0],
            "boolean": [False, False, True],
            "numeric": [1, 2, 3],
        },
        dtype="object",
    ),
)
def test_success_complete_redshift(
    batch_for_datasource: Batch, expectation: gxe.ExpectColumnValuesToBeInTypeList
) -> None:
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    result_dict = result.to_json_dict()["result"]

    assert result.success
    assert isinstance(result_dict, dict)
    assert isinstance(result_dict["observed_value"], str)
    assert isinstance(expectation.type_list, list)
    # We require the user to use DECIMAL, the redshift official type:
    # https://docs.aws.amazon.com/redshift/latest/dg/c_Supported_data_types.html
    # However, we see the string numeric in the sqlalchemy type output.
    if result_dict["observed_value"].lower() != "numeric":
        assert result_dict["observed_value"] in expectation.type_list
    else:
        assert "DECIMAL" in expectation.type_list



from typing import List

from tests.integration.backend_dependencies import BackendDependencies
from tests.integration.integration_test_fixture import IntegrationTestFixture

abs_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []

partition_data = [
    IntegrationTestFixture(
        name="azure_pandas_by_path",
        user_flow_script="tests/integration/test_definitions/abs/select_batch_by_path.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.AZURE],
    ),
    IntegrationTestFixture(
        name="partition_data_on_datetime_azure_pandas",
        user_flow_script="tests/integration/test_definitions/abs/partitioned_on_datetime.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.AZURE],
    ),
]

sample_data: List[IntegrationTestFixture] = []

fluent_datasources = [
    IntegrationTestFixture(
        name="how_to_connect_to_data_on_azure_blob_storage_using_pandas",
        user_flow_script="docs/docusaurus/docs/oss/guides/connecting_to_your_data/fluent/filesystem/how_to_connect_to_data_on_azure_blob_storage_using_pandas.py",
        data_context_dir="tests/integration/fixtures/no_datasources/great_expectations",
        backend_dependencies=[BackendDependencies.AZURE],
    ),
]

abs_integration_tests += connecting_to_your_data
abs_integration_tests += partition_data
abs_integration_tests += sample_data
abs_integration_tests += fluent_datasources



import pathlib
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Mapping, Optional, Union

import pandas as pd
import pytest

from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.data_asset.path.spark.csv_asset import CSVAsset
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.execution_engine import SparkDFExecutionEngine
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark
    from great_expectations.compatibility.pyspark import types as pyspark_types


@dataclass(frozen=True)
class SparkFilesystemCsvDatasourceTestConfig(DataSourceTestConfig):
    # see "read" options: https://spark.apache.org/docs/3.5.3/sql-data-sources-csv.html#data-source-option
    read_options: dict[str, Any] = field(default_factory=dict)
    # see "write" options: https://spark.apache.org/docs/3.5.3/sql-data-sources-csv.html#data-source-option
    write_options: dict[str, Any] = field(default_factory=dict)

    @property
    @override
    def label(self) -> str:
        return "spark-filesystem-csv"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.spark

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        assert not extra_data, "extra_data is not supported for this data source yet."

        tmp_path = request.getfixturevalue("tmp_path")
        assert isinstance(tmp_path, pathlib.Path)

        return SparkFilesystemCsvBatchTestSetup(
            data=data,
            config=self,
            base_dir=tmp_path,
            context=context,
        )


class SparkFilesystemCsvBatchTestSetup(
    BatchTestSetup[SparkFilesystemCsvDatasourceTestConfig, CSVAsset]
):
    def __init__(
        self,
        config: SparkFilesystemCsvDatasourceTestConfig,
        data: pd.DataFrame,
        base_dir: pathlib.Path,
        context: AbstractDataContext,
    ) -> None:
        super().__init__(config=config, data=data, context=context)
        self._base_dir = base_dir

    @property
    def _spark_session(self) -> "pyspark.SparkSession":
        return SparkDFExecutionEngine.get_or_create_spark_session()

    @property
    def _spark_schema(self) -> Union["pyspark_types.StructType", None]:
        from great_expectations.compatibility.pyspark import types as pyspark_types

        column_types = self.config.column_types or {}
        struct_fields = [
            pyspark_types.StructField(column_name, column_type())
            for column_name, column_type in column_types.items()
        ]
        return pyspark_types.StructType(struct_fields) if struct_fields else None

    @property
    def _spark_data(self) -> "pyspark.DataFrame":
        if self._spark_schema:
            return self._spark_session.createDataFrame(self.data, schema=self._spark_schema)
        else:
            return self._spark_session.createDataFrame(self.data)

    @override
    def make_asset(self) -> CSVAsset:
        infer_schema = self._spark_schema is None
        return self.context.data_sources.add_spark_filesystem(
            name=self._random_resource_name(), base_directory=self._base_dir
        ).add_csv_asset(
            name=self._random_resource_name(),
            spark_schema=self._spark_schema,
            header=True,
            infer_schema=infer_schema,
            **self.config.read_options,
        )

    @override
    def make_batch(self) -> Batch:
        return (
            self.make_asset()
            .add_batch_definition_path(name=self._random_resource_name(), path=self.csv_path)
            .get_batch()
        )

    @override
    def setup(self) -> None:
        file_path = self._base_dir / self.csv_path
        self._spark_data.write.format("csv").option("header", True).options(
            **self.config.write_options
        ).save(str(file_path))

    @override
    def teardown(self) -> None: ...

    @property
    def csv_path(self) -> pathlib.Path:
        return pathlib.Path("data.csv")



from __future__ import annotations

import logging

import pytest

from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.compatibility.pydantic import ValidationError
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent import BatchRequest, PandasDatasource
from great_expectations.datasource.fluent.interfaces import (
    DataAsset,
    Datasource,
    HeadData,
)
from great_expectations.exceptions.exceptions import DataContextError
from great_expectations.execution_engine import ExecutionEngine
from great_expectations.validator.computed_metric import MetricValue
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,
    MetricConfigurationID,
)
from great_expectations.validator.metrics_calculator import MetricsCalculator
from tests.expectations.test_util import get_table_columns_metric

logger = logging.getLogger(__name__)


def run_checkpoint_and_data_doc(
    datasource_test_data: tuple[AbstractDataContext, Datasource, DataAsset, BatchRequest],
):
    # context, datasource, asset, batch_request
    context, datasource, asset, batch_request = datasource_test_data

    # Define an expectation suite
    suite_name = "my_suite"
    try:
        context.suites.add(ExpectationSuite(name=suite_name))
    except DataContextError:
        ...

    context.suites.get(name=suite_name)
    validator = context.get_validator(
        batch_request=batch_request,
        expectation_suite_name=suite_name,
    )
    validator.expect_table_row_count_to_be_between(0, 10000)
    validator.expect_column_max_to_be_between(column="passenger_count", min_value=1, max_value=7)
    validator.expect_column_median_to_be_between(column="passenger_count", min_value=1, max_value=4)

    suite = validator.expectation_suite
    suite.save()
    batch_def = asset.add_batch_definition(name="my_batch_definition")

    # Configure and run a checkpoint
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="my_validation_definition", suite=suite, data=batch_def)
    )
    metadata = validator.active_batch.metadata  # type: ignore[union-attr] # active_batch could be None
    if isinstance(datasource, PandasDatasource):
        checkpoint_name = "single_batch_checkpoint"
    else:
        checkpoint_name = (
            f"batch_with_year_{metadata['year']}_month_{metadata['month']}_{suite_name}"
        )
    checkpoint = Checkpoint(
        name=checkpoint_name,
        validation_definitions=[validation_definition],
    )
    checkpoint_result = checkpoint.run()

    # Verify checkpoint runs successfully
    assert checkpoint_result.success, "Running expectation suite failed"
    number_of_runs = len(checkpoint_result.run_results)
    assert number_of_runs == 1, f"{number_of_runs} runs were done when we only expected 1"

    # Grab the validation result and verify it is correct
    validation_result = checkpoint_result.run_results[list(checkpoint_result.run_results.keys())[0]]
    assert validation_result.success

    expected_metric_values = {
        "expect_table_row_count_to_be_between": {
            "value": 10000,
            "rendered_template": "Must have greater than or equal to $min_value and less than or equal to $max_value rows.",  # noqa: E501 # FIXME CoP
        },
        "expect_column_max_to_be_between": {
            "value": 6,
            "rendered_template": "$column maximum value must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
        "expect_column_median_to_be_between": {
            "value": 1,
            "rendered_template": "$column median must be greater than or equal to $min_value and less than or equal to $max_value.",  # noqa: E501 # FIXME CoP
        },
    }
    assert len(validation_result.results) == 3

    for r in validation_result.results:
        assert r.success
        assert r.expectation_config
        assert (
            r.result["observed_value"] == expected_metric_values[r.expectation_config.type]["value"]
        )

        assert r.rendered_content is None
        assert r.expectation_config.rendered_content is None

    # Rudimentary test for data doc generation
    docs_dict = context.build_data_docs()
    assert "local_site" in docs_dict, "build_data_docs returned dictionary has changed"
    assert docs_dict["local_site"].startswith("file://"), (
        "build_data_docs returns file path in unexpected form"
    )
    path = docs_dict["local_site"][7:]
    with open(path) as f:
        data_doc_index = f.read()

    # Checking for ge-success-icon tests the result table was generated and it was populated with a successful run.  # noqa: E501 # FIXME CoP
    assert "ge-success-icon" in data_doc_index
    assert "ge-failed-icon" not in data_doc_index


def run_batch_head(  # noqa: C901 # FIXME CoP
    datasource_test_data: tuple[AbstractDataContext, Datasource, DataAsset, BatchRequest],
    fetch_all: bool | str,
    n_rows: int | float | str | None,  # noqa: PYI041 # FIXME CoP
    success: bool,
) -> None:
    _, datasource, _, batch_request = datasource_test_data
    batch = datasource.get_batch(batch_request=batch_request)

    if success:
        assert n_rows is None or isinstance(n_rows, int)
        assert isinstance(fetch_all, bool)

        execution_engine: ExecutionEngine = batch.data.execution_engine
        execution_engine.batch_manager.load_batch_list(batch_list=[batch])

        metrics: dict[MetricConfigurationID, MetricValue] = {}

        table_columns_metric: MetricConfiguration
        results: dict[MetricConfigurationID, MetricValue]

        table_columns_metric, results = get_table_columns_metric(execution_engine=execution_engine)
        metrics.update(results)

        metrics_calculator = MetricsCalculator(execution_engine=execution_engine)

        # define metric for the total number of rows
        row_count_metric = MetricConfiguration(
            metric_name="table.row_count",
            metric_domain_kwargs={"batch_id": batch.id},
            metric_value_kwargs=None,
        )
        total_row_count: int = metrics_calculator.get_metric(metric=row_count_metric)

        # get the expected column names
        expected_columns: set[str] = set(metrics_calculator.columns())

        head_data: HeadData
        if n_rows is not None and not fetch_all:
            if n_rows < 0 or abs(n_rows) > total_row_count:
                n_rows = 0

            head_data = batch.head(n_rows=n_rows, fetch_all=fetch_all)
            assert isinstance(head_data, HeadData)

            head_data_row_count: int = len(head_data.data.index)

            # if n_rows is between 0 and the total_row_count, that's how many rows we expect
            if 0 <= n_rows <= total_row_count:
                assert head_data_row_count == n_rows
            # if n_rows is greater than the total_row_count, we only expect total_row_count rows
            elif n_rows > total_row_count:
                assert head_data_row_count == total_row_count
            # if n_rows is negative and abs(n_rows) is larger than total_row_count we expect zero rows  # noqa: E501 # FIXME CoP
            elif n_rows < 0 and abs(n_rows) > total_row_count:
                assert head_data_row_count == 0
            # if n_rows is negative, we expect all but the final abs(n_rows)
            else:
                assert head_data_row_count == n_rows + total_row_count
        elif fetch_all:
            total_row_count = metrics_calculator.get_metric(metric=row_count_metric)
            if n_rows:
                head_data = batch.head(n_rows=n_rows, fetch_all=fetch_all)
            else:
                head_data = batch.head(fetch_all=fetch_all)
            assert isinstance(head_data, HeadData)
            assert len(head_data.data.index) == total_row_count
        else:
            # default to 5 rows
            head_data = batch.head(fetch_all=fetch_all)
            assert isinstance(head_data, HeadData)
            assert len(head_data.data.index) == 5

        assert set(metrics[table_columns_metric.id]) == expected_columns  # type: ignore[arg-type] # FIXME CoP

    else:
        with pytest.raises(ValidationError) as e:
            batch.head(n_rows=n_rows, fetch_all=fetch_all)  # type: ignore[arg-type] # FIXME CoP
        n_rows_validation_error = (
            "1 validation error for Head\n"
            "n_rows\n"
            "  value is not a valid integer (type=type_error.integer)"
        )
        fetch_all_validation_error = (
            "1 validation error for Head\n"
            "fetch_all\n"
            "  value is not a valid boolean (type=value_error.strictbool)"
        )
        assert n_rows_validation_error in str(e.value) or fetch_all_validation_error in str(e.value)



from __future__ import annotations

import json
import uuid
from typing import TYPE_CHECKING

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.data_context_key import DataContextKey, StringKey
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store.store import Store
from great_expectations.data_context.store.tuple_store_backend import TupleStoreBackend
from great_expectations.data_context.types.resource_identifiers import (
    GXCloudIdentifier,
)

if TYPE_CHECKING:
    from great_expectations.core.validation_definition import ValidationDefinition


class ValidationDefinitionStore(Store):
    _key_class = StringKey

    def __init__(
        self,
        store_backend: dict | None = None,
        runtime_environment: dict | None = None,
        store_name: str = "no_store_name",
    ) -> None:
        store_backend_class = self._determine_store_backend_class(store_backend)
        if store_backend and issubclass(store_backend_class, TupleStoreBackend):
            store_backend["filepath_suffix"] = store_backend.get("filepath_suffix", ".json")

        super().__init__(
            store_backend=store_backend,
            runtime_environment=runtime_environment,
            store_name=store_name,
        )

    def get_key(self, name: str, id: str | None = None) -> GXCloudIdentifier | StringKey:
        """Given a name and optional ID, build the correct key for use in the ValidationDefinitionStore."""  # noqa: E501 # FIXME CoP
        if self.cloud_mode:
            return GXCloudIdentifier(
                resource_type=GXCloudRESTResource.VALIDATION_DEFINITION,
                id=id,
                resource_name=name,
            )
        return StringKey(key=name)

    @override
    @staticmethod
    def gx_cloud_response_json_to_object_dict(response_json: dict) -> dict:
        response_data = response_json["data"]

        validation_data: dict
        if isinstance(response_data, list):
            if len(response_data) != 1:
                if len(response_data) == 0:
                    msg = f"Cannot parse empty data from GX Cloud payload: {response_json}"
                else:
                    msg = f"Cannot parse multiple items from GX Cloud payload: {response_json}"
                raise ValueError(msg)
            validation_data = response_data[0]
        else:
            validation_data = response_data

        return validation_data

    @override
    @staticmethod
    def _convert_raw_json_to_object_dict(data: dict) -> dict:
        return data

    @override
    def serialize(self, value):
        # In order to enable the custom json_encoders in ValidationDefinition, we need to set `models_as_dict` off  # noqa: E501 # FIXME CoP
        # Ref: https://docs.pydantic.dev/1.10/usage/exporting_models/#serialising-self-reference-or-other-models
        output = value.json(models_as_dict=False, indent=2, sort_keys=True)

        if self.cloud_mode:
            output_dict = json.loads(output)
            output_dict.pop("id", None)
            return output_dict
        else:
            return output

    @override
    def deserialize(self, value):
        from great_expectations.core.validation_definition import ValidationDefinition

        if self.cloud_mode:
            return ValidationDefinition.parse_obj(value)

        return ValidationDefinition.parse_raw(value)

    @override
    def _add(self, key: DataContextKey, value: ValidationDefinition, **kwargs):
        if not self.cloud_mode:
            # this logic should move to the store backend, but is implemented here for now
            value.id = str(uuid.uuid4())
        return super()._add(key=key, value=value, **kwargs)



from __future__ import annotations

import logging
from functools import wraps
from typing import TYPE_CHECKING, Callable, Dict, Optional, Tuple, Type, TypeVar, Union

from typing_extensions import ParamSpec

import great_expectations.exceptions as gx_exceptions
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.metric_function_types import (
    MetricFunctionTypes,
    MetricPartialFunctionTypes,
)
from great_expectations.execution_engine import ExecutionEngine
from great_expectations.expectations.metrics import MetaMetricProvider
from great_expectations.expectations.registry import (
    get_metric_provider,
    register_metric,
    register_renderer,
)
from great_expectations.validator.metric_configuration import MetricConfiguration

if TYPE_CHECKING:
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )

logger = logging.getLogger(__name__)


P = ParamSpec("P")
T = TypeVar("T")


def metric_value(
    engine: Type[ExecutionEngine],
    metric_fn_type: Union[str, MetricFunctionTypes] = MetricFunctionTypes.VALUE,
    **kwargs,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator used to register a specific function as a metric value function.

    Metric value functions are used by MetricProviders to immediately return
    the value of the requested metric.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/expectations/custom_expectations_lp

    Args:
        engine: the *type* of ExecutionEngine that this partial supports
        metric_fn_type: the type of metric function. Usually the default value should be maintained.

    Returns:
        Decorated function
    """

    def wrapper(metric_fn: Callable[P, T]) -> Callable[P, T]:
        @wraps(metric_fn)
        def inner_func(*args: P.args, **kwargs: P.kwargs):
            return metric_fn(*args, **kwargs)

        inner_func.metric_engine = engine  # type: ignore[attr-defined] # FIXME CoP
        inner_func.metric_fn_type = MetricFunctionTypes(metric_fn_type)  # type: ignore[attr-defined] # FIXME CoP
        inner_func.metric_definition_kwargs = kwargs  # type: ignore[attr-defined] # FIXME CoP
        return inner_func

    return wrapper


def metric_partial(
    engine: Type[ExecutionEngine],
    partial_fn_type: MetricPartialFunctionTypes,
    domain_type: Union[str, MetricDomainTypes],
    **kwargs,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator used to register a specific function as a metric partial.

    Metric partial functions are used by MetricProviders to support batching of
    requests for multiple metrics in an ExecutionEngine. Instead of returning
    the metric value immediately, they return a different function that the
    ExecutionEngine can execute locally on your data to obtain the metric value.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/expectations/features_custom_expectations/how_to_add_spark_support_for_an_expectation
        - https://docs.greatexpectations.io/docs/guides/expectations/features_custom_expectations/how_to_add_sqlalchemy_support_for_an_expectation

    Args:
        engine: the *type* of ExecutionEngine that this partial supports
        partial_fn_type: the type of partial function
        domain_type: the type of domain this metric function processes

    Returns:
        Decorated function
    """

    def wrapper(metric_fn: Callable[P, T]) -> Callable[P, T]:
        @wraps(metric_fn)
        def inner_func(*args: P.args, **kwargs: P.kwargs):
            return metric_fn(*args, **kwargs)

        inner_func.metric_engine = engine  # type: ignore[attr-defined] # FIXME CoP
        inner_func.metric_fn_type = MetricPartialFunctionTypes(  # type: ignore[attr-defined] # FIXME CoP
            partial_fn_type
        )  # raises ValueError if unknown type
        inner_func.domain_type = MetricDomainTypes(domain_type)  # type: ignore[attr-defined] # FIXME CoP
        inner_func.metric_definition_kwargs = kwargs  # type: ignore[attr-defined] # FIXME CoP
        return inner_func

    return wrapper


class MetricProvider(metaclass=MetaMetricProvider):
    """Base class for all metric providers.

    MetricProvider classes *must* have the following attributes set:
        1. `metric_name`: the name to use. Metric Name must be globally unique in
           a great_expectations installation.
        2. `domain_keys`: a tuple of the *keys* used to determine the domain of the
           metric
        3. `value_keys`: a tuple of the *keys* used to determine the value of
           the metric.

    In some cases, subclasses of Expectation, such as TableMetricProvider will already
    have correct values that may simply be inherited.

    MetricProvider classes *must* implement the following:
        1. `_get_evaluation_dependencies`. Note that often, _get_evaluation_dependencies should
        augment dependencies provided by a parent class; consider calling super()._get_evaluation_dependencies

    In some cases, subclasses of Expectation, such as MapMetricProvider will already
    have correct implementations that may simply be inherited.

    Additionally, they *may* provide implementations of:
        1. Data Docs rendering methods decorated with the @renderer decorator. See the guide
        "How to create renderers for custom expectations" for more information.

    """  # noqa: E501 # FIXME CoP

    domain_keys: Tuple[str, ...] = tuple()
    value_keys: Tuple[str, ...] = tuple()
    default_kwarg_values: dict = {}

    @classmethod
    def _register_metric_functions(cls) -> None:  # noqa: C901 #  too complex
        metric_name = getattr(cls, "metric_name", None)
        if not metric_name:
            # No metric name has been defined
            return

        metric_domain_keys = cls.domain_keys
        metric_value_keys = cls.value_keys

        for attr_name in dir(cls):
            attr_obj = getattr(cls, attr_name)
            if not (hasattr(attr_obj, "metric_engine") or hasattr(attr_obj, "_renderer_type")):
                # This is not a metric or renderer.
                continue

            if engine := getattr(attr_obj, "metric_engine", None):
                if not issubclass(engine, ExecutionEngine):
                    raise ValueError("metric functions must be defined with an Execution Engine")  # noqa: TRY003 # FIXME CoP

                metric_fn = attr_obj
                metric_definition_kwargs = getattr(metric_fn, "metric_definition_kwargs", {})
                declared_metric_name = metric_name + metric_definition_kwargs.get(
                    "metric_name_suffix", ""
                )
                metric_fn_type: Optional[Union[MetricFunctionTypes, MetricPartialFunctionTypes]] = (
                    getattr(metric_fn, "metric_fn_type", MetricFunctionTypes.VALUE)
                )

                if not metric_fn_type:
                    # This is not a metric (valid metrics possess exectly one metric function).
                    return

                """
                Basic metric implementations (defined by specifying "metric_name" class variable in "metric_class") use
                either "@metric_value" decorator (with default "metric_fn_type" set to "MetricFunctionTypes.VALUE"); or
                "@metric_partial" decorator with specification "partial_fn_type=MetricPartialFunctionTypes.AGGREGATE_FN"
                (which ultimately sets "metric_fn_type" of inner function to this value); or "@column_aggregate_value"
                decorator (with default "metric_fn_type" set to "MetricFunctionTypes.VALUE"); or (applicable for column
                domain metrics only) "column_aggregate_partial" decorator with "partial_fn_type" explicitly set to
                "MetricPartialFunctionTypes.AGGREGATE_FN".  When "metric_fn_type" of metric implementation function is
                of "aggregate partial" type ("MetricPartialFunctionTypes.AGGREGATE_FN"), underlying backend (e.g., SQL
                or Spark) employs "deferred execution" (gather computation needs to build execution plan, then execute
                all computations combined).  Deferred aggregate function calls are bundled (applies to SQL and Spark).
                To instruct "ExecutionEngine" accordingly, original metric is registered with its "declared" name, but
                with "metric_provider" function omitted (set to "None"), and additional "AGGREGATE_FN" metric, with its
                "metric_provider" set to (decorated) implementation function, defined in metric class, is registered.
                Then "AGGREGATE_FN" metric is specified with key "metric_partial_fn" as evaluation metric dependency.
                By convention, aggregate partial metric implementation functions return three-valued tuple, containing
                deferred execution metric implementation function of corresponding "ExecutionEngine" backend (called
                "metric_aggregate") as well as "compute_domain_kwargs" and "accessor_domain_kwargs", which are relevant
                for bundled computation and result access, respectively.  When "ExecutionEngine.resolve_metrics()" finds
                no "metric_provider" (metric_fn being "None"), it then obtains this three-valued tuple from dictionary
                of "resolved_metric_dependencies_by_metric_name" using previously declared "metric_partial_fn" key (as
                described above), composes full metric execution configuration structure, and adds this configuration
                to list of metrics to be resolved as one bundle (specifics pertaining to "ExecutionEngine" subclasses).
                """  # noqa: E501 # FIXME CoP
                if metric_fn_type not in [
                    MetricFunctionTypes.VALUE,
                    MetricPartialFunctionTypes.AGGREGATE_FN,
                ]:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        f"""Basic metric implementations (defined by specifying "metric_name" class variable) only \
support "{MetricFunctionTypes.VALUE.value}" and "{MetricPartialFunctionTypes.AGGREGATE_FN.value}" for "metric_value" \
"metric_fn_type" property."""  # noqa: E501 # FIXME CoP
                    )

                if metric_fn_type == MetricFunctionTypes.VALUE:
                    register_metric(
                        metric_name=declared_metric_name,
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=metric_fn,
                        metric_fn_type=metric_fn_type,
                    )
                else:
                    register_metric(
                        metric_name=f"{declared_metric_name}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=metric_fn,
                        metric_fn_type=metric_fn_type,
                    )
                    register_metric(
                        metric_name=declared_metric_name,
                        metric_domain_keys=metric_domain_keys,
                        metric_value_keys=metric_value_keys,
                        execution_engine=engine,
                        metric_class=cls,
                        metric_provider=None,
                        metric_fn_type=metric_fn_type,
                    )
            elif hasattr(attr_obj, "_renderer_type"):
                register_renderer(object_name=metric_name, parent_class=cls, renderer_fn=attr_obj)

    @classmethod
    def get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        """This should return a dictionary:

        {
          "dependency_name": MetricConfiguration,
          ...
        }
        """
        return (
            cls._get_evaluation_dependencies(
                metric=metric,
                configuration=configuration,
                execution_engine=execution_engine,
                runtime_configuration=runtime_configuration,
            )
            or {}
        )

    @classmethod
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        dependencies: Dict[str, MetricConfiguration] = {}

        if execution_engine is None:
            return dependencies

        try:
            metric_name: str = metric.metric_name
            _ = get_metric_provider(
                f"{metric_name}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                execution_engine,
            )
            dependencies["metric_partial_fn"] = MetricConfiguration(
                metric_name=f"{metric_name}.{MetricPartialFunctionTypes.AGGREGATE_FN.metric_suffix}",
                metric_domain_kwargs=metric.metric_domain_kwargs,
                metric_value_kwargs=metric.metric_value_kwargs,
            )
        except gx_exceptions.MetricProviderError:
            pass

        return dependencies



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Literal, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    MulticolumnMapExpectation,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.expectations.model_field_descriptions import (
    COLUMN_LIST_DESCRIPTION,
    IGNORE_ROW_IF_DESCRIPTION,
    MOSTLY_DESCRIPTION,
)
from great_expectations.render import RenderedStringTemplateContent
from great_expectations.render.components import LegacyRendererType
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    substitute_none_for_missing,
)

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

logger = logging.getLogger(__name__)

EXPECTATION_SHORT_DESCRIPTION = (
    "Expect that the sum of row values in a specified column list "
    "is the same for each row, and equal to a specified sum total."
)
SUM_TOTAL_DESCRIPTION = "Expected sum of columns"
DATA_QUALITY_ISSUES = [DataQualityIssues.NUMERIC.value]
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]


class ExpectMulticolumnSumToEqual(MulticolumnMapExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectMulticolumnSumToEqual is a \
    Multicolumn Map Expectation.

    Multicolumn Map Expectations are evaluated for a set of columns and ask a yes/no question about the row-wise relationship between those columns.
    Based on the result, they then calculate the percentage of rows that gave a positive answer.
    If the percentage is high enough, the Expectation considers that data valid.

    Args:
        column_list (tuple or list): {COLUMN_LIST_DESCRIPTION}
        sum_total (int or float): {SUM_TOTAL_DESCRIPTION}

    Other Parameters:
        ignore_row_if (str): \
            "both_values_are_missing", "either_value_is_missing", "neither" \
            {IGNORE_ROW_IF_DESCRIPTION} Default "neither".
        mostly (None or a float between 0 and 1): \
            {MOSTLY_DESCRIPTION} \
            For more detail, see [mostly](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#mostly). Default 1.
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2   test3
            0 	1       2       4
            1 	2       -2       7
            2 	4   	4       -3

    Code Examples:
        Passing Case:
            Input:
                ExpectMulticolumnSumToEqual(
                    column_list=["test", "test2", "test3"],
                    sum_total=7,
                    mostly=0.66
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 1,
                    "unexpected_percent": 33.33333333333333,
                    "partial_unexpected_list": [
                      {{
                        "test": 4,
                        "test2": 4,
                        "test3": -3
                      }}
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 33.33333333333333,
                    "unexpected_percent_nonmissing": 33.33333333333333
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectMulticolumnSumToEqual(
                    column_list=["test", "test2", "test3"],
                    sum_total=7
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "element_count": 3,
                    "unexpected_count": 1,
                    "unexpected_percent": 33.33333333333333,
                    "partial_unexpected_list": [
                      {{
                        "test": 4,
                        "test2": 4,
                        "test3": -3
                      }}
                    ],
                    "missing_count": 0,
                    "missing_percent": 0.0,
                    "unexpected_percent_total": 33.33333333333333,
                    "unexpected_percent_nonmissing": 33.33333333333333
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    sum_total: Union[float, SuiteParameterDict] = pydantic.Field(description=SUM_TOTAL_DESCRIPTION)
    ignore_row_if: Union[
        Literal["all_values_are_missing", "any_value_is_missing", "never"],
        SuiteParameterDict,
    ] = pydantic.Field(
        default="all_values_are_missing",
        description=IGNORE_ROW_IF_DESCRIPTION,
    )

    # This dictionary contains metadata for display in the public gallery
    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": [
            "core expectation",
            "multi-column expectation",
        ],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    map_metric = "multicolumn_sum.equal"
    success_keys = ("mostly", "sum_total")
    args_keys = (
        "column_list",
        "sum_total",
    )

    class Config:
        title = "Expect multicolumn sum to equal"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[ExpectMulticolumnSumToEqual]) -> None:
            MulticolumnMapExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column_list", RendererValueType.ARRAY),
            ("sum_total", RendererValueType.NUMBER),
            ("mostly", RendererValueType.NUMBER),
            ("ignore_row_if", RendererValueType.STRING),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        template_str = ""
        if params.column_list:
            array_param_name = "column_list"
            param_prefix = "column_list_"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            template_str += "Sum across columns " + cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

        if params.mostly and params.mostly.value < 1.0:
            renderer_configuration = cls._add_mostly_pct_param(
                renderer_configuration=renderer_configuration
            )
            template_str += " must be $sum_total, at least $mostly_pct % of the time."
        else:
            template_str += " must be $sum_total."

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> List[RenderedStringTemplateContent]:
        runtime_configuration = runtime_configuration or {}
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            ["column_list", "sum_total", "mostly"],
        )
        if params["mostly"] is not None:
            params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
        mostly_str = "" if params.get("mostly") is None else ", at least $mostly_pct % of the time"
        sum_total = params.get("sum_total")  # noqa: F841 # FIXME CoP

        column_list_str = ""
        for idx in range(len(params["column_list"]) - 1):
            column_list_str += f"$column_list_{idx!s}, "
            params[f"column_list_{idx!s}"] = params["column_list"][idx]
        last_idx = len(params["column_list"]) - 1
        column_list_str += f"$column_list_{last_idx!s}"
        params[f"column_list_{last_idx!s}"] = params["column_list"][last_idx]
        template_str = f"Sum across columns {column_list_str} must be $sum_total{mostly_str}."
        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]
